/* This file is part of the KMPlayer project
 *
 * Copyright (C) 2005 Koos Vriezen <koos.vriezen@xs4all.nl>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Steet, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */
#include <config.h>
#include <math.h>
#include <sys/types.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include <gtk/gtk.h>
#define DBUS_API_SUBJECT_TO_CHANGE
#include <dbus/dbus.h>

#include "kmplayer.h"
#include "kmplayerprocess.h"
#ifdef __ARMEL__
# include "npplayer.h"
# include "kmplayer_lists.h"
#endif

using namespace KMPlayer;

static void dbusSend (DBusConnection *conn,
        const char *dest, const char *path, const char *iface,
        const char *func, int first_arg_type, ...) {
    va_list var_args;
    DBusMessage *msg = dbus_message_new_method_call (
            dest,
            path,
            iface,
            func);
    if (first_arg_type != DBUS_TYPE_INVALID) {
        va_start (var_args, first_arg_type);
        dbus_message_append_args_valist (msg, first_arg_type, var_args);
        va_end (var_args);
    }
    dbus_message_set_no_reply (msg, TRUE);
    dbus_connection_send (conn, msg, NULL);
    dbus_connection_flush (conn);
    dbus_message_unref (msg);
}

//#define dbusSend(conn, dest, path, iface, func, first_arg_type, rest...) \
    osso_rpc_run((osso_context_t*)conn, dest, path, iface, func, NULL, first_arg_type, ## rest);

Process::Process (ProcessNotify * ctrl, const char * n, bool xv)
 : m_notify (ctrl), m_viewer (0L), m_name (n),
      m_state (NotRunning), m_old_state (NotRunning),// m_process (0L),
      start_timer (0), no_video (false), has_xv (xv), is_local_file (false)
      {}

Process::~Process () {
    stop ();
    if (start_timer)
        g_source_remove (start_timer);
}

//WId Process::widget () {
//    return 0;
//}

KMPLAYER_NO_EXPORT bool Process::playing () const {
    return m_state > Ready;
}

//void Process::setAudioLang (int, const QString &) {}

//void Process::setSubtitle (int, const QString &) {}

KMPLAYER_NO_EXPORT void Process::setPosition (int p) {
    m_position = p;
    m_notify->setPosition (m_position);
}

KMPLAYER_NO_EXPORT bool Process::pause () {
    return false;
}

KMPLAYER_NO_EXPORT bool Process::seek (int /*pos*/, bool /*absolute*/) {
    return false;
}

KMPLAYER_NO_EXPORT bool Process::volume (int /*pos*/, bool /*absolute*/) {
    return false;
}

KMPLAYER_NO_EXPORT bool Process::saturation (int /*pos*/, bool /*absolute*/) {
    return false;
}

KMPLAYER_NO_EXPORT bool Process::hue (int /*pos*/, bool /*absolute*/) {
    return false;
}

KMPLAYER_NO_EXPORT bool Process::contrast (int /*pos*/, bool /*absolute*/) {
    return false;
}

KMPLAYER_NO_EXPORT bool Process::brightness (int /*pos*/, bool /*absolute*/) {
    return false;
}

KMPLAYER_NO_EXPORT void Process::setVideoWindow (int, int, int, int) {}

//bool Process::grabPicture (const KURL & /*url*/, int /*pos*/) {
//    return false;
//}

//bool Process::supports (const char * source) const {
//    for (const char ** s = m_supported_sources; s[0]; ++s) {
//        if (!strcmp (s[0], source))
//            return true;
//    }
//    return false;
//}

KMPLAYER_NO_EXPORT bool Process::stop () {
    if (!playing ()) return true;
    return false;
}

KMPLAYER_NO_EXPORT bool Process::quit () {
    debugLog() << "quit" << endl;
    stop ();
    setState (NotRunning);
    return !playing ();
}

static gboolean cb_scheduledUpdateAspects (void * p) {
    static_cast <Process *> (p)->updateAspects ();
    return false; // single shot
}

KMPLAYER_NO_EXPORT void Process::updateAspects () {
    if (m_notify) {
        float aspect = m_aspect;
        if (aspect < 0.01)
            aspect = m_height > 0 ? 1.0 * m_width / m_height : 0;
        m_notify->setAspect (aspect);
    }
}

static gboolean cb_scheduledStateChanged (void * p) {
    static_cast <Process *> (p)->scheduledStateChanged ();
    return false; // single shot
}

KMPLAYER_NO_EXPORT void Process::setState (State newstate) {
    if (m_state != newstate) {
        m_old_state = m_state;
        m_state = newstate;
        start_timer = g_timeout_add (0, cb_scheduledStateChanged, static_cast <void *> (this));
    }
}

KMPLAYER_NO_EXPORT void Process::scheduledStateChanged () {
    m_notify->stateChanged (this, m_old_state, m_state);
    start_timer = 0;
}

KMPLAYER_NO_EXPORT bool Process::play (NodePtr _mrl) {
    m_mrl = _mrl;
    Mrl *m = mrl ();
    m_url = m ? m->absolutePath () : String ();
    no_video = m ? m->audio_only : false;
    m_position = m_length = 0;
    m_user_stop_position = 0;
    m_width = m_height = 0;
    m_aspect = 0.0;
    is_local_file = false;
    return false;
}

KMPLAYER_NO_EXPORT bool Process::ready (unsigned int viewer) {
    m_viewer = viewer;
    setState (Ready);
    return true;
}

KMPLAYER_NO_EXPORT unsigned int Process::viewer () const {
    return m_viewer;
}

KMPLAYER_NO_EXPORT void Process::infoMessage (const String &msg) {
    if (m_mrl)
        m_mrl->document ()->message (MsgInfoString, (void *) &msg);
}

void Process::setUserStop () {
    if (m_state > Buffering && m_length > 0 && is_local_file)
        m_user_stop_position = m_position;
}

//-----------------------------------------------------------------------------

static void watchProcessFunc (GPid pid, gint status, gpointer data) {
#ifdef DEBUG_PROCESS_WATCH
    debugLog () << " watchProcessFunc " << endl;
#endif
    g_spawn_close_pid (pid);
    static_cast <UnixProcess *> (data)->processExited (pid);
}

static gboolean watchMPlayerStdin (GIOChannel *, GIOCondition, gpointer data) {
    static_cast <UnixProcess *> (data)->writeEvent ();
#ifdef DEBUG_PROCESS_WATCH
    debugLog () << " watchProcessStdin " << endl;
#endif
    return true;
}

static void setupProcessFunc (gpointer data) {
}

static gboolean
watchProcessStdout (GIOChannel *src, GIOCondition, gpointer data) {
#ifdef DEBUG_PROCESS_WATCH
    debugLog () << " watchProcessStdout " << endl;
#endif
    if (G_IO_IN)
        static_cast <UnixProcess *> (data)->readEvent (src);
    return true;
}

UnixProcess::UnixProcess (UnixProcessListener *l, int flags)
    : listener (l), io_flags (flags),
    pin (0L), pout (0L), win (0), wout (0) {
}

bool UnixProcess::start (const String &pwd, const String &cmdline) {
    int argc;
    gchar **argv;
    GError *err = 0L;
    int fdin, fdout, fderr;
    eof = false;
    debugLog () << cmdline << endl;
    if (g_shell_parse_argv ((const char *) cmdline, &argc, &argv, &err) &&
            g_spawn_async_with_pipes ((const char *) pwd,
                argv,
                0L,
                (GSpawnFlags)(G_SPAWN_SEARCH_PATH | G_SPAWN_DO_NOT_REAP_CHILD),
                setupProcessFunc,
                static_cast <void*> (this),
                &process_id,
                &fdin,
                &fdout,
                0L /*&fderr*/,
                &err)) {
        //darn glib: assertion err!=null g_error_free (err);
        g_strfreev (argv);
        g_child_watch_add(process_id, watchProcessFunc,static_cast<void*>(this));
        pin = g_io_channel_unix_new (fdin);
        g_io_channel_set_encoding (pin, NULL, NULL);
        g_io_channel_set_buffered (pin, false);
        GIOFlags flags = g_io_channel_get_flags (pin);
        g_io_channel_set_flags (pin, (GIOFlags)(flags|G_IO_FLAG_NONBLOCK), 0L);

        pout = g_io_channel_unix_new (fdout);
        g_io_channel_set_encoding (pout, NULL, NULL);
        flags = g_io_channel_get_flags (pout);
        g_io_channel_set_flags (pout, (GIOFlags)(flags|G_IO_FLAG_NONBLOCK), 0L);
        g_io_channel_set_buffer_size (pout, 0);
        wout = g_io_add_watch (pout, G_IO_IN, watchProcessStdout, this);
        return true;
    }
    errorLog () << err->message << endl;
    g_error_free (err);
    return false;
}

void UnixProcess::stop () {
    closeStdIn ();
    closeStdOut ();
    if (process_id) {
        ::kill (process_id, SIGINT);
        for (int n = 0; process_id && n < 100; n++) {
            usleep (10000); // 10ms
            while (g_main_context_iteration (0L, false))
                if (!process_id)
                    break;
        }
        if (process_id) {
            errorLog () << "interupt process" << endl;
            ::kill (process_id, SIGTERM);
            for (int n = 0; process_id && n < 50; n++) {
                usleep (10000); // 10ms
                while (g_main_context_iteration (0L, false))
                    if (!process_id)
                        break;
            }
        }
        if (process_id)
            processExited (process_id); // give up
    }
}

bool UnixProcess::running () const {
    return process_id;
}

void UnixProcess::pollStdIn (bool b) {
    if (b) {
        if (!win)
            win = g_io_add_watch (pin, G_IO_OUT, watchMPlayerStdin, this);
    } else if (win) {
        g_source_remove (win);
        win = 0;
    }
}

int UnixProcess::readStdOut (char *buf, int len) {
    gsize nr = 0;;
    if (pout) {
        GIOStatus status = g_io_channel_read_chars (pout, buf, len, &nr, 0L);
        if (status == G_IO_STATUS_EOF || status == G_IO_STATUS_ERROR /*FIXME*/)
            eof = true;
    } else {
        eof = true;
    }
    return nr;
}

int UnixProcess::writeStdIn (const char *data, int len) {
    gsize nr = 0;
    g_io_channel_write_chars (pin, data, len, &nr, 0L);
    //g_io_channel_flush (pin, 0L);
    return nr;
}

void UnixProcess::closeStdIn () {
    pollStdIn (false);
    if (pin) {
        g_io_channel_shutdown (pin, true, 0L);
        g_io_channel_unref (pin);
        pin = 0L;
    }
}

void UnixProcess::closeStdOut () {
    if (wout) {
        g_source_remove (wout);
        wout = 0;
    }
    if (pout) {
        g_io_channel_shutdown (pout, true, 0L);
        g_io_channel_unref (pout);
        pout = 0L;
    }
}

void UnixProcess::readEvent (GIOChannel *channel) {
    listener->readEvent (this);
}

void UnixProcess::writeEvent () {
    if (pin)
        listener->writeEvent (this);
}

void UnixProcess::processExited (GPid pid) {
    if (pid == process_id) {
        process_id = 0;
        closeStdIn ();
        listener->processExited (this);
    }
}

//-----------------------------------------------------------------------------

#define NO_SEEK 0x7fffffff

MPlayer::MPlayer (ProcessNotify *ctrl, const char *nm)
 : Process (ctrl, nm, true),
   unix_process (this, UnixProcess::StdIn | UnixProcess::StdOut),
   path ("mplayer"),
   audio_driver (NULL),
   video_driver (NULL),
   wait_for_pause (false) {}

MPlayer::~MPlayer () {
}

KMPLAYER_NO_EXPORT bool MPlayer::play (NodePtr node) {
    debugLog () << "MPlayer::play" << endl;
    if (playing ())
        return pause ();
    wait_for_pause = false;
    setState (Buffering);
    Process::play (node);
    if (m_url.isEmpty ()) {
        setState (Ready);
        return false;
    }
    request_seek = NO_SEEK;
    String wd (getenv ("HOME"));
    URL url (m_url);
    is_local_file = url.isLocalFile ();
    if (is_local_file) {
        m_url = url.path ();
        wd = url.path ();
        int pos = wd.lastIndexOf (String ("/"));
        if (pos > -1)
            wd.truncate (pos);
    }
    request_quit = false;
    outbuf[0] = 0;
    String cmdline (path);
    cmdline += " -nofs -slave -framedrop -identify -vo ";
    char buf[128];
    //if (has_xv)
    //    snprintf (buf, sizeof(buf), "xv -wid %u ", m_viewer);
    //else
        snprintf (buf, sizeof(buf),
                video_driver
                ? video_driver
                : "xv");
    cmdline += String (buf);
    if (audio_driver) {
        snprintf (buf, sizeof (buf), " -ao %s", audio_driver);
        cmdline += String (buf);
    }
    snprintf (buf, sizeof(buf), " -wid %u ", m_viewer);
    cmdline += String (buf);
    Mrl *mrl = this->mrl ();
    if (mrl->repeat > 0)
        cmdline += String (" -loop ") + String::number (mrl->repeat);
    if (is_local_file && mrl->clip_begin > 100)
        cmdline += String (" -ss ") + String::number ((int) mrl->clip_begin/100);
    cmdline += Char (' ');

    const TrieString aspect_ratio ("aspectRatio");
    bool found_attr = false;
    for (Node *n = mrl; n && !found_attr; n = n->parentNode ()) {
        if (n->id != id_node_group_node && n->id != id_node_playlist_item)
            continue;
        Attribute *a = static_cast <Element *> (n)->attributes ().first ();
        for (; a; a = a->nextSibling ()) {
            if (aspect_ratio == a->name ()) {
                String asp = a->value ();
                if (asp.isEmpty ())
                    continue;
                int p = asp.indexOf (Char (':'));
                if (p > 0) {
                    int a = asp.left (p).toInt ();
                    int b = asp.mid (p + 1).toInt ();
                    if (a > 0 && b > 0)
                        cmdline += String ("-aspect ") +
                            String::number (a) + Char (':') + String::number (b);
                } else {
                    float a = (float) asp.toDouble ();
                    if (a > 0.01) {
                        char buf[32];
                        snprintf (buf, sizeof (buf), "%0.3f", a);
                        cmdline += String ("-aspect ") +
                            String (buf).replace (Char (','), Char ('.'));
                    }
                }
                cmdline += Char (' ');
                found_attr = true;
            } else if (a->name () == StringPool::attr_panzoom) {
                String coord[4];
                if (Mrl::parsePanZoomString (a->value (), coord)) {
                    cmdline += String ("-vf crop=") +
                        String::number (coord[2].toInt()) + Char (':') +
                        String::number (coord[3].toInt()) + Char (':') +
                        String::number (coord[0].toInt()) + Char (':') +
                        String::number (coord[1].toInt());
                }
                cmdline += Char (' ');
                found_attr = true;
            }
        }
    }

    char *quoted = g_shell_quote ((const char *) m_url);
    cmdline += String (quoted);
    g_free (quoted);
    if (unix_process.start (wd, cmdline)) {
        m_mrl = node;
        setState (Process::Buffering); //FIXME
        return true;
    }
    setState (Ready);
    return false;
}

KMPLAYER_NO_EXPORT bool MPlayer::pause () {
    if (state () != Paused)
        wait_for_pause = !wait_for_pause;
    for (StringList::iterator i = commands.begin (); i != commands.end (); ++i)
        if (i->startsWith ("pause")) {
            commands.erase (i);
            return true;
        }
    return sendCommand (String ("pause"));
}

KMPLAYER_NO_EXPORT bool MPlayer::ready (unsigned int viewer) {
    if (m_state < Ready)
        return Process::ready (viewer);
    return false;
}

KMPLAYER_NO_EXPORT bool MPlayer::stop () {
    debugLog() << "stop mplayer" << endl;
    return sendCommand (String ("quit"));
}

KMPLAYER_NO_EXPORT bool MPlayer::quit () {
    if (playing ()) {
        request_quit = true;
        stop ();
        while (g_main_context_iteration (0L, false))
            if (!unix_process.running ())
                break;
        for (int n = 0; unix_process.running () && n < 50; n++) {
            usleep (10000); // 10ms
            while (g_main_context_iteration (0L, false))
                if (!unix_process.running ())
                    break;
        }
        if (unix_process.running ()) {
            errorLog () << "stop MPlayer" << endl;
            unix_process.stop ();
        }
        commands.clear ();
        request_quit = false;
    }
    setState (NotRunning);
    return true;
}

KMPLAYER_NO_EXPORT bool MPlayer::seek (int pos, bool absolute) {
    if (/*request_seek != NO_SEEK &&*/ commands.size () > 1) {
        StringList::iterator i = commands.begin ();
        StringList::iterator end ( commands.end () );
        for (++i; i != end; ++i)
            if ((*i).startsWith (String ("seek"))) {
                i = commands.erase (i);
                request_seek = NO_SEEK;
                break;
            }
    }
    if (request_seek != NO_SEEK)
        return false;
    if (absolute && m_length > 0) {
        pos -= m_position;
        absolute = false;
    }
    request_seek = pos;
    char buf[32];
    snprintf (buf, sizeof (buf), "seek %d %d", pos/10, absolute ? 2 : 0);
    return sendCommand (String (buf));
}

KMPLAYER_NO_EXPORT bool MPlayer::volume (int pos, bool /*absolute*/) {
    return sendCommand (String ("volume ") + String::number (pos));
}

KMPLAYER_NO_EXPORT bool MPlayer::saturation (int pos, bool absolute) {
    return true;
}

KMPLAYER_NO_EXPORT bool MPlayer::hue (int pos, bool absolute) {
    return true;
}

KMPLAYER_NO_EXPORT bool MPlayer::contrast (int pos, bool absolute) {
    return true;
}

KMPLAYER_NO_EXPORT bool MPlayer::brightness (int pos, bool absolute) {
    return true;
}

KMPLAYER_NO_EXPORT void MPlayer::processExited (UnixProcess *) {
    unix_process.closeStdOut ();
    commands.clear ();
    if (m_state > Ready && !request_quit)
        setState (Ready);
}

KMPLAYER_NO_EXPORT void MPlayer::writeEvent (UnixProcess *) {
    String data;
    if (command_chunk.isEmpty ()) {
        if (commands.size ()) {
            data = *commands.begin ();
            commands.pop_front ();
        }
    } else
        data = command_chunk;
    debugLog () << "eval " << data /*<< endl*/;
    gsize nr = 0;
    if (!data.isEmpty ())
        nr = unix_process.writeStdIn ((const char*) data, data.length ());
    if (nr < data.length ()) {
        //debugLog () << "partial " << nr << " " << data.length() << endl;
        command_chunk = String (((const char*) data) + nr);
    } else if (!commands.size ()) {
        // FIXME: always remove and readd when new data from stdin arrives
        unix_process.pollStdIn (false);
    }
}

KMPLAYER_NO_EXPORT void MPlayer::readEvent (UnixProcess *) {
    int inbuf = strlen (outbuf);
    gsize nr = unix_process.readStdOut (outbuf + inbuf, sizeof (outbuf) -inbuf);
    //debugLog() << "read " << nr << " chars had " << inbuf << endl;
    outbuf[inbuf + nr] = 0;
    char *tok = outbuf;
    for (int i = 0; i < inbuf + nr; i++) {
        if (outbuf[i] == '\r' && outbuf[i+1] != '\n') {
            outbuf[i] = 0;
            char * c = strstr (tok, "A:");
            if (!c)
                c = strstr (tok, "V:");
            if (c) {
                setPosition ((int)(10 * strtof (c + 2, 0L)));
                request_seek = NO_SEEK;
                setState (Playing);
            } else {
                c = strstr (tok, "Cache fill:");
                if (c) {
                    char *p = strchr (tok + 11, '%');
                    if (p)
                        *p = 0;
                    //debugLog() << "c " << (c + 11) << endl;
                    m_notify->setLoading ((int)strtof (c + 11, 0L));
                }
                if (m_state == Paused)
                    setState (Playing);
            }
            tok = outbuf + i + 1;
        } else if (outbuf[i] == '\n') {
            bool handled = false;
            outbuf[i] = 0;
            debugLog() << "> " << tok << endl;
            if (m_state == Buffering) {
                char * p = strstr (tok, "Start");
                if (p && strstr (p, "play")) { // find a 'Start.*play' pattern
                    m_notify->setLoading (99);
                    handled = true;
                } else {
                    char * c = strstr (tok, "Cache fill:");
                    if (c) {
                        char *p = strchr (tok + 11, '%');
                        if (p)
                            *p = 0;
                        //debugLog() << "c2 " << (c + 11) << endl;
                        m_notify->setLoading ((int)strtof (c + 11, 0L));
                        handled = true;
                    } else {
                        char * v = strstr (tok, "Video: no video");
                        if (v) {
                            setHasVideo (false);
                            handled = true;
                        }
                    }
                }
            }
            if (!handled) {
                if (!strncmp (tok, "ID_", 3)) {
                    handled = true;
                    if (!strncmp (tok + 3, "LENGTH", 6)) {
                        m_length = (int)(10 * strtof (tok + 10, 0L));
                        m_notify->setLength (m_length);
                    } else if (!strncmp (tok + 3, "PAUSED", 6)) {
                        if (wait_for_pause) {
                            wait_for_pause = false;
                            setState (Paused);
                        } // else ignore, a second 'pause' cmd should follow
                    } else if (!strncmp (tok + 3, "VIDEO_WIDTH", 11)) {
                        m_width = (int) strtof (tok + 15, 0L);
                        if (m_height > 0)
                            updateAspects ();
                    } else if (!strncmp (tok + 3, "VIDEO_HEIGHT", 12)) {
                        m_height = (int) strtof (tok + 16, 0L);
                        if (m_width > 0)
                            updateAspects ();
                    } else if (!strncmp (tok + 3, "VIDEO_ASPECT", 12)) {
                        // convert in C locale
                        m_aspect = (float) String (tok + 16).toDouble ();
                        if (m_aspect > 0.01)
                            updateAspects ();
                    }
                } else {
                    char *icy = strstr (tok, "ICY Info");
                    if (icy) {
                        //debugLog () << "icy found" << endl;
                        char *t = strstr (icy + 8, "StreamTitle=");
                        if (t) {
                            char *e = strchr (t + 12, ';');
                            if (e)
                                *e = 0;
                            infoMessage (String (t + 12));
                        }
                        handled = true;
                    } else {
                        // Crop: 320 x 200, 0 ; 20
                        char *crop = strstr (tok, "Crop:");
                        if (crop) {
                            char *by = strchr (tok + 5, 'x');
                            if (by) {
                                char *endptr = NULL;
                                int w = strtol (tok + 5, &endptr, 10);
                                if (w > 0 && endptr != tok + 5) {
                                    int h = strtol (by + 1, &endptr, 10);
                                    if (h > 0 && endptr != by + 1) {
                                        m_width = w;
                                        m_height = h;
                                        updateAspects ();
                                        handled = true;
                                    }
                                }
                            }
                        }
                    }
                }
            }
            tok = outbuf + i + 1;
        }
    }
    inbuf = tok - outbuf;
    if (inbuf > 0)
        memmove (outbuf, tok, inbuf);
    outbuf[inbuf] = 0;
}

KMPLAYER_NO_EXPORT bool MPlayer::sendCommand (const String & cmd) {
    if (playing ()) {
        commands.push_back (cmd + Char ('\n'));
        unix_process.pollStdIn (true);
        return true;
    }
    return false;
}

#ifdef __ARMEL__

static NSNotify ns_notify;

extern "C" {

static void npSetDimension (void *ndata, int w, int h) {
    NpPlayer *player = (NpPlayer *) ndata;
    player->setWidth (w);
    player->setHeight (h);
    player->updateAspects ();
}

static void npGetUrl (void *inst, uint32_t stream,
        const char *url, const char *target,
        unsigned plen, const char *post, void *ndata) {
    ((NpPlayer *) ndata)->requestStream (inst, stream, url, target, plen, post);
}

static char *nsEvaluate (const char *script, void *ndata) {
    String result = ((NpPlayer *) ndata)->evaluateScript (script);
    if (result.isEmpty ())
        return g_strdup ("");
    return g_strdup ((const char *)result);
}

static void npFinish (void *inst, uint32_t stream, void *ndata) {
    ((NpPlayer *) ndata)->destroyStream (inst, stream);
}

static gint streamCompare (gconstpointer a, gconstpointer b) {
    return (long)a - (long)b;
}

static gboolean deleteStreams (gpointer key, gpointer value, gpointer data) {
    (void) key;
    delete (NpStream *)value;
    return false; // continue
}

static gboolean copyParams (gpointer key, gpointer value, gpointer data) {
    char ***tuple = (char ***)data;
    tuple[0][0] = (char *)key;
    tuple[1][0] = (char *)value;
    tuple[0]++;
    tuple[1]++;
    return false; // continue
}

static gboolean openStreams (gpointer key, gpointer value, gpointer data) {
    int active = (int)(long)data;
    NpStream *ns = (NpStream *)value;
    (void) key;
    if (ns->finish_reason == NpStream::NoReason && !ns->job) {
        ++active;
        ns->open ();
    }
    return active > 4;
}

static gboolean findClosedStreams (gpointer key, gpointer value, gpointer data) {
    NpStream *ns = (NpStream *)value;
    (void) key;
    if (ns->finish_reason > NpStream::NoReason) {
        *(NpStream **)data = ns;
        return true;
    }
    return false;
}

} // extern "C"

KDE_NO_CDTOR_EXPORT NpPlayer::NpPlayer (ProcessNotify * ctrl)
 : Process (ctrl, "npp", false),
   streams (NULL),
   mime (NULL),
   npdata (NULL),
   argc (0),
   argn (NULL),
   argv (NULL),
   processing(false) {
    plugins = g_tree_new ((GCompareFunc)::strcmp);
    ns_notify.setDimension = npSetDimension;
    ns_notify.getUrl = npGetUrl;
    ns_notify.evaluate = nsEvaluate;
    ns_notify.finishStream = npFinish;
}

KDE_NO_CDTOR_EXPORT NpPlayer::~NpPlayer () {
    if (mime)
        g_free (mime);
}

bool NpPlayer::isFlashMimeType (const String &mime)
{
    return mime == "application/x-shockwave-flash" ||
        mime == "application/futuresplash";
}

KMPLAYER_NO_EXPORT bool NpPlayer::ready (unsigned int viewer) {
    debugLog () << "NpPlayer::ready" << endl;
    m_viewer = viewer;
    if (m_state < Ready) {
        // only flash for now
        mime = g_strdup ("application/x-shockwave-flash");
        const char *lib = "/usr/lib/browser/plugins/libflashplayer.so";
        NPPluginLib *nplib = (NPPluginLib *) g_tree_search (plugins,
                (GCompareFunc)::strcmp, mime);
        if (!nplib) {
            nplib = nppPluginInit (lib);
            if (!nplib) {
                warningLog() << "failed to initialize plugin lib" << endl;
                return false;
            }
            g_tree_insert (plugins, (char *) mime, nplib);
        }
        nppInstanceWindowParent (npdata, (void *)(long)viewer);
        setState (Ready);
        return true;
    }
    return false;
}

KMPLAYER_NO_EXPORT bool NpPlayer::play (NodePtr node) {
    if (playing ())
        return true;
    setState (Buffering);
    Process::play (node);
    if (m_url.isEmpty ()) {
        setState (Ready);
        return false;
    }
    nppInstanceTopUrl (m_url);
    streams = g_tree_new (streamCompare);

    GTree *params = g_tree_new ((GCompareFunc)::strcmp);
    Mrl *mrl = this->mrl ();
    debugLog () << "NpPlayer::play " << mrl->mimetype << endl;
    if (mrl->id == id_node_html_object) {
        for (Node *n = mrl->firstChild (); n; n = n->nextSibling ()) {
            /*if (n->id == KMPlayer::id_node_param) {
                KMPlayer::Element *e = static_cast <KMPlayer::Element *> (n);
                String name = e->getAttribute (KMPlayer::StringPool::attr_name);
                if (!g_tree_lookup (params, (const char *) name))
                    g_tree_insert (params,
                            strdup ((const char *) name),
                            strdup ((const char *) e->getAttribute (KMPlayer::StringPool::attr_value)));
            } else*/ if (n->id == KMPlayer::id_node_html_embed) {
                KMPlayer::Element *e = static_cast <KMPlayer::Element *> (n);
                Attribute *a = e->attributes ().first ();
                for (; a; a = a->nextSibling ()) {
                    String nm = a->name().toString();
                    if (!g_tree_lookup (params, (const char *) nm))
                        g_tree_insert (params,
                                strdup ((const char *) nm),
                                strdup ((const char *) a->value ()));
                }
            }
        }
    } else if (NpPlayer::isFlashMimeType (mrl->mimetype)) {
        g_tree_insert (params, strdup ("SRC"), strdup (m_url));
    } else {
        String flashvars = String ("file=") + URL::encode_string (m_url);
        m_url = "http://video.linuxfoundation.org/sites/all/modules/custom/os_video_player/mediaplayer.swf";
        g_tree_insert (params, strdup ("SRC"), strdup (m_url));
        g_tree_insert (params, strdup ("flashvars"), strdup (flashvars));
    }
    argc = g_tree_nnodes (params);
    argn = (char **) malloc (argc * sizeof (char *));
    argv = (char **) malloc (argc * sizeof (char *));
    char **tuple[2] = { argn, argv };
    g_tree_traverse (params, copyParams, G_IN_ORDER, tuple);
    g_tree_destroy (params);

    npdata = nppInstanceOpen (mime, argc, argn, argv, (void *)this, &ns_notify);
    if (!npdata) {
        g_tree_destroy (streams);
        streams = NULL;
        setState (Ready);
        warningLog() << "failed to create new instance" << endl;
        return false;
    }
    return true;
}

KMPLAYER_NO_EXPORT void NpPlayer::pluginEmbedded () {
    if (state () == Buffering) {
        nppInstanceStart ((const char *)m_url, mime, npdata);
        setState (Playing);
    }
}

KMPLAYER_NO_EXPORT void NpPlayer::requestStream (void *inst, uint32_t sid,
        const String & url, const String & target,
        unsigned post_len, const char *post) {
    URL uri (m_url, url);
    debugLog () << "NpPlayer::request " << sid << " '" << uri << "'" << endl;
    if (!target.isEmpty ()) {
        debugLog () << "new page request " << target << endl;
        if (url.startsWith ("javascript:")) {
            String result = evaluateScript (url.mid (11));
            debugLog() << "result is " << result << endl;
            if (result == "undefined")
                uri = URL ();
            else
                uri = URL (m_url, result); // probably wrong ..
        }
        /*if (!uri.url ().isEmpty ())
            emit openUrl (uri, target);*/
        nppStreamFinished (sid, NpStream::BecauseDone, this);
    } else if (url.startsWith ("javascript:")) {
        String result = evaluateScript (url.mid (11));
        const char *data = (const char *)result;
        nppStreamData (sid, data, strlen (data), this);
        nppStreamFinished (sid, NpStream::BecauseDone, this);
    } else {
        NpStream * ns = new NpStream (this, sid, uri, post_len, post);
        g_tree_insert (streams, (gpointer) sid, ns);
        if (url != uri.url ())
            nppStreamRedirected (sid, (const char *)uri.url ());
        processStreams ();
    }
}

KMPLAYER_NO_EXPORT void NpPlayer::destroyStream (void *inst, uint32_t sid) {
    NpStream * ns = (NpStream *) g_tree_lookup (streams, (gpointer) sid);
    if (ns) {
        nppStreamFinished (sid, NpStream::BecauseStopped, this);
        g_tree_remove (streams, (gpointer) sid);
        delete (ns);
    }
}

KMPLAYER_NO_EXPORT bool NpPlayer::stop () {
    if (npdata) {
        g_tree_foreach (streams, deleteStreams, NULL);
        g_tree_destroy (streams);
        streams = NULL;
        nppInstanceClose (npdata);
        npdata = NULL;
        setState (Ready);
    }
    for (int i = 0; i < argc; i++) {
        free (argn[i]);
        free (argv[i]);
    }
    if (argn) {
        free (argn);
        free (argv);
        argn = argv = NULL;
    }
    argc = 0;
    return true;
}

KMPLAYER_NO_EXPORT bool NpPlayer::quit () {
    stop ();
    if (mime)
        g_free (mime);
    mime = NULL;
    nppInstanceWindowClose (npdata); //FIXME
    setState (NotRunning);
    return true;
}

KMPLAYER_NO_EXPORT
void NpPlayer::setVideoWindow (int x, int y, int w, int h) {
    //debugLog() << "NpPlayer::setVideoWindow " << w << "x" << h << endl;
    if (!(x < 0 || y < 0 || w <= 0 || h <= 1 /* hack !!*/ ||
                x > 800 || y > 480 || w > 800 || h > 480))
        nppInstanceWindowDimension (npdata, x, y, w, h);
}

KMPLAYER_NO_EXPORT void NpPlayer::processStreams () {
    if (!processing) {
        processing = true;
        streams_deleted = 0;
        int active = 0;
        g_tree_traverse (streams, openStreams, G_IN_ORDER, &active);
        while (streams_deleted > 0) {
            NpStream *ns = NULL;
            g_tree_traverse (streams, findClosedStreams, G_IN_ORDER, &ns);
            if (ns) {
                g_tree_remove (streams, (gpointer) ns->stream);
                delete ns;
            } else {
                errorLog() << "findClosedStreams didn't find a stream" << endl;
                break;
            }
            streams_deleted--;
        }
        processing = false;
    }
}

KMPLAYER_NO_EXPORT void NpPlayer::streamFinished (NpStream *ns) {
    nppStreamFinished (ns->stream, ns->finish_reason, this);
    if (!processing) {
        g_tree_remove (streams, (gpointer) ns->stream);
        delete ns;
        processStreams ();
    } else {
        streams_deleted++;
    }
}

KMPLAYER_NO_EXPORT String NpPlayer::evaluateScript (const String &script) {
    debugLog() << "evaluateScript " << script << endl;
    if (script == "window.location+\"__flashplugin_unique__\"" ||
            script == "top.location+\"__flashplugin_unique__\"")
        return "http://www.example.com/__flashplugin_unique__";
    return "undefined";
}

KDE_NO_CDTOR_EXPORT NpStream::NpStream (NpPlayer *p, uint32_t id, const URL &u,
        unsigned post_len, const char *post)
 : player (p), stream (id), url (u), job (NULL), finish_reason (NoReason) {
    if (post_len && post) {
        post_data.resize (post_len);
        memcpy (post_data.data (), post, post_len);
    }
}

KDE_NO_CDTOR_EXPORT NpStream::~NpStream () {
    close ();
}

KMPLAYER_NO_EXPORT void NpStream::open () {
    debugLog () << "NpStream::open " << url.url () << endl;
    if (url.url().startsWith ("javascript:")) {
        String result = player->evaluateScript (url.url().mid (11));
        const char *data = (const char *)result;
        nppStreamData (stream, data, strlen (data), NULL /*FIXME*/);
        finish_reason = BecauseDone;
        player->streamFinished (this);
    } else {
        bytes = 0;
        job = asyncGet (this, url.url ());
        if (post_data.size ()) {
            String buf;
            int data_pos = -1;
            bool nl = true;
            for (int i = 0; i < post_data.size () && data_pos < 0; ++i) {
                char c = post_data.data () [i];
                switch (c) {
                    case '\r':
                        break;
                    case '\n':
                        if (nl)
                            data_pos = i + 1;
                        else
                            job->setHeader (buf);
                        buf.truncate (0);
                        nl = true;
                        break;
                    default:
                        nl = false;
                        buf += Char (c);
                }
            }
            job->setHeader ("Expect:");
            if (data_pos > 0)
                job->setHttpPostData (ByteArray (post_data.data () + data_pos,
                            post_data.size () - data_pos));
        }
        job->start ();
    }
}

KMPLAYER_NO_EXPORT void NpStream::close () {
    if (job)
        job->kill ();
    if (NoReason == finish_reason)
        nppStreamFinished (stream, BecauseStopped, player);
    job = NULL;
}

KMPLAYER_NO_EXPORT void NpStream::jobData (IOJob *jb, ByteArray & data) {
    //debugLog () << "NpStream::jobData " << url.url () << endl;
    if (!bytes)
        nppStreamInfo (stream, jb->contentType (), jb->contentLength ());
    bytes += data.size ();
    nppStreamData (stream, data.data (), data.size (), NULL /*FIXME*/);
}

KMPLAYER_NO_EXPORT void NpStream::jobResult (IOJob *jb) {
    debugLog () << "NpStream::jobResult " << bytes << endl;
    finish_reason = jb->error () || !bytes ? BecauseError : BecauseDone;
    job = NULL;
    player->streamFinished (this);
}

KMPLAYER_NO_EXPORT void NpStream::redirected (IOJob *, const String & uri) {
    url = URL (url, uri);
    nppStreamRedirected (stream, (const char *)url.url ());
}

#endif //__ARMEL__

Download::Download (NodePtr &d, ProcessNotify *n, int _fd)
    : Element (d, id_node_download),
      job (NULL), notify (n),
      size (0), bytes (0), fd (_fd)
{}

Download::~Download ()
{
    if (job) {
        job->kill ();
        ::close (fd);
    }
}

void Download::closed ()
{
    file = getAttribute ("file");
    uri = getAttribute (StringPool::attr_src);
    title = getAttribute (StringPool::attr_title);
    Element::closed ();
}

void Download::begin ()
{
    setState (state_began);
    bytes = 0;
    size = 0;
    debugLog() << "Download " << uri << " as " << file << " cap:" << title << endl;
    job = asyncGet (this, uri);
    if (fd > -1) {
        job->start ();
    } else {
        String sz = getAttribute ("size");
        if (sz.isEmpty ()) {
            fd = ::open (file, O_WRONLY | O_CREAT | O_TRUNC, 0664);
        } else {
            size = strtoll (sz, NULL, 10);
            fd = ::open (file, O_WRONLY | O_APPEND, 0664);
            bytes = ::lseek (fd, 0, SEEK_END);
        }
        if (fd > 0)
            job->start (bytes);
    }
    notify->updateDownload (this, 0);
}

void Download::jobData (IOJob *jb, ByteArray &data)
{
    int old_perc = size > 0 ? int (100.0 * bytes / size) : 0;
    if (!bytes) {
        size = jb->contentLength ();
        if (size > 0)
            setAttribute ("size", String::number (size));
    }
    bytes += data.size ();
    int new_perc = size > 0 ? int (100.0 * bytes / size) : 0;
    if (new_perc >= 100)
        new_perc = 99;
    if (old_perc != new_perc)
        notify->updateDownload (this, new_perc);
    if (data.size ()) {
        int nr = write (fd, data.data (), data.size ());
        if (nr < data.size ()) {
            String err ("Couldn't write data");
            if (nr < 0)
                err += Char (' ') + String (strerror (errno));
            jb->kill (false);
            notify->errorMsg (err);
        }
    }
}

void Download::jobResult (IOJob *jb)
{
    if (fd > 0)
        close (fd);
    fd = -1;
    job = NULL;
    if (!jb->error ()) {
        notify->updateDownload (this, 100);
        NodePtr safe (this);
        parentNode ()->removeChild (this); // deletes this
        notify->updateDownloads ();
    }
}

void Download::redirected (IOJob * job, const String & uri)
{
}

/*
String saved_file;
IOJob *job;
ProcessNotify *notify;
unsigned long size;
unsigned long bytes;
int fd;*/

Downloader::Downloader (ProcessNotify *ctrl, Application *app)
 : Process (ctrl, "downloader", false),
   application (app) {
}

Downloader::~Downloader () {
}

static String genFileName (const String &str, bool skip_on_invalid) {
    String name;
    bool valid = true;
    for (const char *s = (const char *)str; s && *s; ++s)
        switch (*s) {
            case '\n':
            case ';':
            case ':':
            case '*':
            case '?':
            case ',':
            case '&':
            case '>':
            case '<':
            case '|':
            case '"':
            case '\'':
                if (!skip_on_invalid)
                    return String ();
                break;
            default:
                name += Char (*s);
        }
    return name;
}

bool KMPLAYER_NO_EXPORT Downloader::play (NodePtr mrl) {
    Process::play (mrl);
    if (!m_url.isEmpty ()) {
        String name, title;
        int pos = m_url.lastIndexOf (Char ('/'));
        if (pos > 0)
            name = genFileName ((const char *) m_url.mid (pos + 1), false);
        PlaylistRole *title_role = (PlaylistRole*) mrl->role (RolePlaylist);
        if (title_role) {
            title = title_role->caption ();
            if (name.isEmpty ())
                name = title;
        }
        if (!name.isEmpty ()) {
            if (title.isEmpty ())
                title = name;
            name = genFileName ((const char *) name, true);
        }
        String path = String (application->save_folder);
        if (!name.isEmpty ())
            path += name;
        gchar *filename = app_file_chooser (application, (const char *) path,
                GTK_FILE_CHOOSER_ACTION_SAVE);
        if (filename) {
            int fd = open (filename, O_WRONLY | O_CREAT | O_TRUNC, 0664);
            if (fd > 0) {
                NodePtr doc = m_notify->downloads ();
                Element *n = new Download (doc, m_notify, fd);
                doc->appendChild (n);
                n->setAttribute (StringPool::attr_src, m_url);
                n->setAttribute (StringPool::attr_title, title);
                n->setAttribute ("file", filename);
                n->closed ();
                n->begin ();
                processNotify ()->updateDownloads ();
            } else {
                String err ("Couldn't open file ");
                err += String (filename) + Char (' ');
                err += strerror (errno);
                processNotify ()->errorMsg (err);
            }
            app_store_save_directory (application, filename);
        }
    }
    setState (NotRunning);
    return false;
}

KMPLAYER_NO_EXPORT bool Downloader::stop () {
    return true;
}
